from .statement import FinancialStatement
from .elements import LineItem, MultiLineItem, Elements, \
    map_to_periods, relocate, Schedule, ScheduleFunction, ScheduleFactory
from .functions import arr
from .functions.base import statfunc
from .optimize import cumsum, mean, prior, total


import numpy as np
import pandas as pd

MATH_METHODS = ['__add__', '__sub__', '__mul__', '__truediv__', '__radd__', '__rsub__', '__rmul__', '__rtruediv__']

### Helpers for converting between pandas period objects and strings ###
PANDAS_FREQ_OFFSET = [
    '10T', 'D', 'B', 'C', 'W', 'SM', 'SMS', 'M', 'BM' , 'CBM', 'MS', 'BMS', 'CBMS', 'Q', 'BQ',
    'QS', 'BQS', 'A', 'Y', 'BA', 'BY', 'AS', 'YS', 'BAS', 'BYS'
]
FREQ_MAP = {freqstr: v for freqstr, v in zip(PANDAS_FREQ_OFFSET, np.arange(len(PANDAS_FREQ_OFFSET)))}
PERIODS_TO_STRF = {
    '10T': '%d-%m-%y %H:%M', 
    'D': '%d-%m-%y',
    'W': '%d-%m-%y',
    'W-SUN': '%d-%b-%y',
    'M': '%Y-%m',
    'Q': '%b-%Y',
    'Q-DEC': '%b-%Y',
    'A-DEC': '%Y',
    'A': '%Y',
}

class GraphNotAssignedError(Exception):
    def __init__(self, msg=None, *args, **kwargs):
        if msg is None:
            msg = 'Object has not been assigned to a statement and so does not have a graph.'
        super().__init__(msg, *args, **kwargs)

class FinStatBaseObject:
    @property
    def _metaparams(self):
        """
        Used to pass metadata as kwargs on _constructor calls.

        Used in _constructor_wrapper
        """
        return {meta.lstrip('_'): getattr(self, meta) for meta in self._metadata}

    @property
    def _metaattrs(self):
        """
        Used to pass metadata as kwargs to `_update_meta` calls
        """
        return {meta: getattr(self, meta) for meta in self._metadata}

    @property
    def _constructor_kwargs(self):
        return {k: getattr(self, k) for k in self._init_keys}

    def _filter_constructor_kwargs(self, kwargs):
        for k, v in self._constructor_kwargs.items():
            if k not in kwargs:
                kwargs[k] = v

        return kwargs

    @property
    def short_name(self):
        return self._short_name

    @property
    def periods(self):
        return self._periods

    @property
    def graph(self):
        return self._graph

    @property
    def G(self):
        return self._graph

    @property
    def has_graph(self):
        return self.graph is not None

    def set_graph(self, graph):
        self._graph = graph

    def update_graph(self, graph):
        from .statement import StatGraph
        super().__setattr__('_graph', StatGraph(self, graph))

    def has_node(self, name):
        return self.has_graph and name in self._graph.nodes

    @property
    def AS_EXPR(self):
        return self.has_graph and self.graph.graph['OPEN_MODEL_CONTEXT']

    def open_model_context(self):
        if self.has_graph:
            self.graph.graph['OPEN_MODEL_CONTEXT'] = True
        else:
            raise GraphNotAssignedError

    def close_model_context(self):
        if self.has_graph:
            self.graph.graph['OPEN_MODEL_CONTEXT'] = False
        else:
            raise GraphNotAssignedError

    @property
    def model_context_is_open(self):
        return self.graph.graph['OPEN_MODEL_CONTEXT']

    @property
    def M(self):
        return self._graph.graph['model_mngr']

    def solve_out_of_context(self):
        self.open_model_context()
        self.M.solve()
        self.close_model_context()

    @property
    def successors(self):
        if self.has_graph:
            return list(self.graph.successors(self.short_name))
        else:
            raise GraphNotAssignedError

    @property
    def is_multi(self):
        NotImplementedError

    @property
    def index_as_frame(self):
        return self.index.to_frame().reset_index(drop=True)

    def _shorten(self, name, short_name=None) -> str:
        """"
        Logic for creating shorthand name for an LineItem object
        
        Examples:
            'Revenue' -> 'revenue'
            'Cost of Goods Sold' -> 'cogs'
            'Amortization' -> 'amort'

        Parameters:
            name: str
        """ 
        if short_name is None:
            if isinstance(name, (int, float)):
                return str(name)
            if name in ['', None]:
                return name
            elif name == 'Income Statement':
                return 'istat'
            elif len(name.split(' ')) > 1:
                return ''.join(s[0].lower() for s in name.split(' '))
            elif len(name) > 10:
                return name[:5].lower()
            else:       
                return name.lower()
        else:
            return short_name

    def to_frame(self):
        return pd.DataFrame(self.values, index=self.index, columns=self.columns)

    def compare_freq(self, freq_a:str, freq_b:str):
        freq_a = freq_a.split('-')[0]
        freq_a = ''.join([i for i in freq_a if not i.isdigit()])
        freq_b = freq_b.split('-')[0]
        freq_b = ''.join([i for i in freq_b if not i.isdigit()])
        return FREQ_MAP[freq_a] > FREQ_MAP[freq_b]



import warnings
import typing as typ
import inspect
import functools as ft
import numpy as np
import pandas as pd

import networkx as nx

from ..base import FinStatBaseObject
from finstat.functions.base import FuncMixin
from .lineitem import MultiLineItem

ScheduleType = typ.TypeVar('ScheduleType', bound='Schedule')

# Helpers map list of value and date items to a schedule containing
# all dates in a given set of periods
def _create_arr(values:np.ndarray, periods:pd.PeriodIndex) ->  np.ndarray:
    """
    Creates m x n array, where m is df.shape[0], or the number of values to be mapped, and n is len(periods), or the number of dates
    available on which values can be mapped. Each row in the array contains only one value in the index == 0 location of that row.
    
    Parameters:
        values: pandas Series or numpy n x 1 ndarray
        periods: pandas PeriodIndex
            
    Return:
        arr: (m x n) numpy ndarray
            m == len(values), or the number of items to be mapped and n == len(periods), or the number of periods considered.
            arr contains one value per row, all in the index 0 location of that row.
    """
    arr = np.zeros((len(values), periods.size))
    arr[:, 0] = values
    
    return arr

def relocate(arr:np.ndarray, locs:np.ndarray, direction:str='forward') -> np.ndarray:
    """
    Intermediate helper that shifts each index 0 location value in `arr` to the location found in `locs`.
    
    Numba could be used to improve speed for large `arr`.
    
    `ogrid` is used to produce arrays of row and column indices. The locations are then subtracted from the
    column indices. `arr` is then resorted with update column indices. This results in the values at index 0 
    moving to the index location specified in `locs`. 

    Parameters:
        values: pandas Series or numpy n x 1 ndarray
        periods: pandas PeriodIndex
            
    Return:
        arr: (m x n) numpy ndarray
            m == len(values), or the number of items to be mapped and n == len(periods), or the number of periods considered.
            arr contains one value per row, each value positioned in the index location specified by `locs`.

    """
    rows, column_indices = np.ogrid[:arr.shape[0], :arr.shape[1]]
    if direction == 'forward':
        column_indices = column_indices - locs[:, np.newaxis]
    elif direction == 'backward':
         column_indices = column_indices + locs[:, np.newaxis]
    return arr[rows, column_indices]

def map_to_periods(values:np.ndarray, dates:np.ndarray, periods:pd.PeriodIndex) -> np.ndarray:
    """
    Process:
        1. Create m x n array for each m item in df to be mapped and n periods on which the values are to be mapped
        2. 
    
    Parameters:
        values: n x 1 np.ndarray
        dates: n x 1 np.ndarray of datetime-like objects
        periods: pandas PeriodIndex

    Return:
        arr: (m x n) numpy ndarray
            m == len(values), or the number of items to be mapped and n == len(periods), or the number of periods considered.
            arr contains one value per row, each value positioned in the index location specified by `locs`.
    """
    arr = _create_arr(values, periods)
    locs = periods.searchsorted(dates)
    arr = relocate(arr, locs)
    
    return arr

class ScheduleFactory(FuncMixin):
    def __init__(self, func:typ.Callable, finstat:'FinancialStatement'=None, **kwargs):
        self.EXPECTED_FACTORS = inspect.getfullargspec(func).annotations
        self.schedfunc = func
        self.factors = {}
        self.finstat = finstat
        self.fkwargs = kwargs

    def set_factors(self):
        if self.finstat is not None:
            for factor in self.EXPECTED_FACTORS:
                self.factors[factor] = getattr(self.finstat, factor)
        else:
            for k, v in self.fkwargs.items():
                if k in self.EXPECTED_FACTORS:
                    self.factors[k] = v # when another schedule is passed don't make it a factor

    def __call__(self, *args, **kwargs):
        self.set_factors()
        sched = self.schedfunc(**self.factors)
        sched = Schedule(sched, *args, factory=self, **kwargs, **self.call_kwargs)
        if self.finstat is not None:
            sched.set_graph(self.finstat.graph)

        return sched

class ScheduleFunction(FuncMixin):
    def __init__(self, sched, func, *args, **kwargs):
        ft.update_wrapper(self, func)
        self.sched = sched
        self.func = func
        self.fargs = args
        
        self._periods = kwargs.pop('periods', None)
        self.fkwargs = kwargs

    @property
    def periods(self):
        if hasattr(self, '_periods') and self._periods is not None:
            return self._periods
        else:
            if self.sched.has_graph:
                return self.graph.graph['periods']
            else:
                raise ValueError('You must provide `periods` argument if the schedule is not part of a FinancialStatement')

    @property
    def index(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", FutureWarning)
            return pd.MultiIndex.from_frame(self.sched)

    @property
    def is_multi(self):
        return True # allows ScheduleFunction to act as a MultiLineItem in `add_account` method

    def col_values(self, sched, col:str):
        return sched[col].values

    def __call__(self, *args, **kwargs):
        sched = self.sched._factory() if self.sched.has_factory else self.sched
        fargs = [self.col_values(sched, col) for col in self.fargs]
        fkwargs = {k: self.col_values(sched, col) for k, col in self.fkwargs.items()}
        obj = self.func(*fargs, periods=self.periods, **fkwargs)

        return MultiLineItem(
            obj,
            columns=self.periods,
            index=pd.MultiIndex.from_frame(sched),
            *args,
            **kwargs,
            **self.call_kwargs
        )

class Schedule(FinStatBaseObject, pd.DataFrame):
    """
    Container and handler class for a list of invoice or purchase orders.
    """
    _metadata = ['_name', '_short_name', '_graph', '_sched_maker']

    def __init__(self, 
        *args, 
        name:str=None, 
        short_name:str=None, 
        graph:nx.DiGraph=None,
        factory:typ.Callable=None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self._graph = graph
        self._name = name
        self._short_name = short_name if short_name else self._shorten(name)
        self._factory = factory

    @property
    def _constructor(self):
        return Schedule

    @property
    def name(self):
        return self._name

    @property
    def sched_maker(self):
        return self._sched_maker

    @property
    def has_factory(self):
        return self._factory is not None

    def set_factory(self, factory):
        self._factory = factory

    def map_schedule(self,
        value_column:str=None,
        date_column:str=None,
        periods:pd.PeriodIndex=None,
        return_sched:bool=False
        ):
        funcobj = ScheduleFunction(self, map_to_periods, value_column, date_column, periods=periods)

        if self.has_graph and not return_sched:
            return funcobj
        else:
            return funcobj()



import functools as ft
import warnings
import typing as typ
import numpy as np

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy
from pandas._typing import TimedeltaConvertibleTypes, TimestampConvertibleTypes
from pandas.core.indexing import _LocIndexer, _iLocIndexer

import networkx as nx   

from ..base import FinStatBaseObject, PERIODS_TO_STRF
from .elements import Elements

def _constructor_wrapper(func, self):
    """
    Takes output from a DataFrameGroupBy method, then 
    reinstantiates the object and passing additional data specific to a
    finstat object 
    """
    @ft.wraps(func)
    def wrapper(*args, **kwargs):
        ret = func(self, *args, **kwargs)
        result = self.obj._constructor(ret.values, index=ret.index, columns=ret.columns, **ret._constructor_kwargs, **kwargs)
        
        return result
    
    return wrapper

class MultiLineGroupBy(DataFrameGroupBy):
    """
    Customized groupby class to control type of the output frame
    """
    def __getattribute__(self, name):
        WRAPPED_FUNCS = ['sum']
        if name in WRAPPED_FUNCS:
            func = super().__getattribute__(name)
            return _constructor_wrapper(func, self)
        else:
            return super().__getattribute__(name)

class MultiLineResampler:
    """
    Customized groupby class to 
    """
    def __init__(self, obj, *args, **kwargs):
        self.obj = obj
        self.resampled = obj.to_frame().T.resample(*args, **kwargs)

    def sum(self, *args, **kwargs):
        self._funcstr = 'sum'
        return self._call(*args, **kwargs)

    def _call(self, *args, **kwargs):
        resamp = getattr(self.resampled, self._funcstr)(*args, **kwargs).T
        return self.obj._constructor(resamp.values, index=resamp.index, columns=resamp.columns, **self.obj._constructor_kwargs)

class IndexerMixin:
    def _check_metric(self):
        if self.obj.has_graph:
            nodetype = self.obj.graph.nodes(data='nodetype')[self.obj.short_name]
            if nodetype == 'metric':
                raise Exception('`loc` and `iloc` __setitem__ not available for metric nodetypes')
        else:
            raise Exception('LineItem must have a `_graph` attribute to use `loc`/`iloc` __setitem__')

    def update_model(self, update_key, value):
        self.obj.M._update_param(self.obj.short_name, update_key, value)
        self.obj.solve_out_of_context()

class StatLoc(IndexerMixin, _LocIndexer):
    def __setitem__(self, key, value):
        self._check_metric()

        if isinstance(key, slice):
            index = self.obj.loc[key].index # not circular b/c it uses __getitem__, not __setitem__
            update_key = index.strftime(PERIODS_TO_STRF[index.freqstr])
        else:
            update_key = key

        self.update_model(update_key, value)
   
        super().__setitem__(key, value) # assign updated ONLY AFTER successful model operations to ensure values are not changed

class StatiLoc(IndexerMixin, _iLocIndexer):
    def __setitem__(self, key, value):
        self._check_metric()

        index = self.obj.index[key]
        update_key = index.strftime(PERIODS_TO_STRF[index.freqstr])

        self.update_model(update_key, value)
   
        super().__setitem__(key, value) # assign updated ONLY AFTER successful model operations to ensure values are not changed

class LineItem(FinStatBaseObject, pd.Series):
    """
    A 1-dimensional time-series node in a FinancialStatment graph. Represents a single line item in a 
    financial statement across the given periods.

    The values of an  LineItem can have a "given" object in that its values are provided externally, as opposed to metric
    """
    _metadata = ['_name', '_short_name', '_periods', '_graph']
    _init_keys = ['name', 'short_name', 'graph']

    def __init__(self, values, 
        name:str=None, 
        short_name:str=None, 
        graph:nx.DiGraph=None, 
        **kwargs
    ): 
        super().__init__(values, name=name, **kwargs)

        self._short_name = None
        if short_name:
            self._short_name = short_name
        elif name is not None:
            if not isinstance(name, (pd.Period, tuple)): # when pd.Period or tuple is passed, this is result of _constructor_sliced call from MultiLineItem
                self._short_name = self._shorten(name)

        self._periods = self.index
        self._graph = graph
        self._name = name

    def __hash__(self):
        return hash(self._name)

    @property
    def _constructor(self):
        return LineItem

    @property
    def is_multi(self):
        return False
 
    def update_periods(self, periods:pd.PeriodIndex):
        self.index = periods
        self._periods = periods

    @property
    def loc(self):
        return StatLoc('loc', self)

    @property
    def iloc(self):
        return StatiLoc('iloc', self)

    def to_series(self):
        return pd.Series(self.values, index=self.index, name=self.name)

    def to_multi(self, index:pd.MultiIndex=None):
        """
        Converts the one-level series to a 
        """
        return MultiLineItem(
            self.values.reshape(1,-1), 
            index=index, 
            columns=self.index,
            **self._constructor_kwargs
        )

    def plot(self, *args, **kwargs):
        return self.to_series().plot(*args, **kwargs)

class MultiLineItem(FinStatBaseObject, pd.DataFrame):
    _metadata = LineItem._metadata
    _init_keys = LineItem._init_keys

    @property
    def _constructor(self):
        return MultiLineItem
    
    @property
    def _constructor_sliced(self):
        return LineItem
    
    def __init__(self, 
        values,
        name:str=None, 
        short_name:str=None, 
        graph:nx.DiGraph=None, 
        index:pd.Index=None,
        **kwargs
    ):  
        index = self._clean_index(index, name)
        super().__init__(values, index=index, **kwargs)
    
        self._name = name
        self._short_name = short_name if short_name else (self._shorten(name) if name else None)
        self._graph = graph
        self._periods = self.columns

    def _clean_index(self, index:pd.Index, name:str):
        """
        Attaches 'LineItem' level to index if it does not already exist

        LineItem is only attached if 'LineItem' is not already in the index and
        the object has not been instantiaed through a call to _constructor
        
        The other approach is to override the various methods with _constructor calls to pass a
        flag that prevents the `_reindex` call.
        """
        if index is not None and name is not None:
            if isinstance(index, pd.Index) and index.name == 'Item':
                index = pd.Index([name], name='Item')
            elif 'Item' in index.names:
                index = index.droplevel('Item')

            if not 'Item' in index.names:
                index = self.insert_item_level_in_index(index, name)

        return index

    def multimath(self, item, funcname, **kwargs):
        if isinstance(item, MultiLineItem): # if both terms are MultiLineItem, must insure the indices are aligned
            accounts = Elements(self, item)
            if not accounts.are_aligned:
                left, right = accounts.align()

                return getattr(left, funcname)(right, **kwargs) # passes right back to `multimath` which will now find accounts.are_aligned == True

        result = getattr(super(), funcname)(item)
        return self._constructor(result.values, index=result.index, columns=result.columns, **kwargs)

    def __add__(self, item, **kwargs):
        return self.multimath(item, '__add__', **kwargs)

    def __radd__(self, item, **kwargs):
        return self.multimath(item, '__radd__', **kwargs)

    def __sub__(self, item, **kwargs):
        return self.multimath(item, '__sub__', **kwargs)

    def __rsub__(self, item, **kwargs):
        return self.multimath(item, '__rsub__', **kwargs)

    def __mul__(self, item, **kwargs):
        return self.multimath(item, '__mul__', **kwargs)

    def __rmul__(self, item, **kwargs):
        return self.multimath(item, '__rmul__', **kwargs)

    def __truediv__(self, item, **kwargs):
        return self.multimath(item, '__truediv__', **kwargs)

    def __rtruediv__(self, item, **kwargs):
        return self.multimath(item, '__rtruediv__', **kwargs)

    @property
    def name(self):
        return self._name

    @property
    def is_multi(self):
        return True

    @property
    def has_single_row(self):
        return self.shape[0] == 1

    def update_periods(self, periods:pd.PeriodIndex):
        self.columns = periods
        self._periods = periods

    def insert_item_level_in_index(self, index, name:str=None):
        name = self.name if name is None else name
        frame = index.to_frame().reset_index(drop=True)
        frame.insert(0, 'Item', name)

        return pd.MultiIndex.from_frame(frame)

    def insert_item_level(self, name:str=None):
        """
        Inserts an 'LineItem' level into the index of the LineItem object passed.

        Parameters:
            account: LineItem
        """
        acct = self.copy()
        idx = self.insert_item_level_in_index(acct.index, name)
        acct.index = idx

        return acct

    def categorize_level(self, name, categories):
        idx_frame = self.index_as_frame
        cats = pd.Categorical(idx_frame[name], categories=categories, ordered=True)
        idx_frame.loc[:, name] = cats
        self.index = pd.MultiIndex.from_frame(idx_frame)

    def groupby(
        self,
        by=None,
        axis=0,
        level:None=None,
        *args, **kwargs
        ):
        """Ripped from pandas groupby"""

        if level is None and by is None:
            raise TypeError("You have to supply one of 'by' and 'level'")

        return MultiLineGroupBy(
            obj=self,
            keys=by,
            axis=self._get_axis_number(axis),
            level=level,
            *args, **kwargs
        )

    def resample(self, *args, **kwargs):
        """
        Resample data structure along the columns / periods axis.

        Unlike LineItem objects, MultiLineItem objects must be rotated via `T` then resampled, then rotated back. This is not
        currently supported by MultiLineItem, so the pd.DataFrame is used and the MultiLineItem reconsituted via
        `_constructor`.
        """
        return MultiLineResampler(self, *args, **kwargs)

    # def sum(self, *args, **kwargs):
    #     result = super().sum()
    #     kwargs = result.set_constructor_kwargs(kwargs)
    #     for k, v in kwargs.items(): # `sum` will not pass any _metaparams, so `set_constructor_kwargs` will be full on None that should be replaced
    #         if v is None:
    #             kwargs[k] = self._metaparams[k]

    #     return result._constructor(result.values, index=result.index, **kwargs)

    # def total(self, item=None, **kwargs):
    #     if 'name' not in kwargs or kwargs['name'] is None:
    #         kwargs['name'] = 'Total ' + self.name
    #         kwargs['short_name'] = 'tot_' + self.short_name
        
    #     index_name =  index_values(self.index, kwargs['name'])
    #     index = pd.MultiIndex.from_tuples([index_name], names=self.index.names)
        
    #     total = self.sum().to_multi(index)

    #     return self._constructor(total.values, index=total.index, columns=total.columns, **self.set_constructor_kwargs(kwargs))

    # def cumsum(self, **kwargs):
    #     cumsum = super().cumsum(axis=1)
    #     return self._constructor(cumsum.values, index=cumsum.index, columns=cumsum.columns, **self.set_constructor_kwargs(kwargs))

    # def where(self, func, comp, if_true=None, if_false=None, if_true_neg:bool=False, if_false_neg:bool=False, **kwargs):
    #     if_true = self if if_true is None else if_true
    #     if_false = comp if if_false is None else if_false

    #     if if_true_neg:
    #         if_true = -if_true
    #     if if_false_neg:
    #         if_false = -if_false

    #     result = np.where(func(self, comp), if_true, if_false)

    #     return self._constructor(result, index=self.index, columns=self.columns, **self.set_constructor_kwargs(kwargs))

    # @property
    # def loc(self):
    #     return LineItemLoc('loc', self)

    # @property
    # def iloc(self):
    #     return LineItemiLoc('iloc', self)

    # def update(self, data:typ.Union[dict, list]):
    #     """
    #     Assigns new values to object and cascades updates
    #     to any successor nodes.

    #     Data to update successors is calculated using the Metric keywords stored
    #     in each successor node.
    #     """
    #     if isinstance(data, dict):
    #         for k, v in data.items():
    #             self.loc(push=True, update=True)[k] = v
    #     elif isinstance(data, pd.DataFrame):
    #         if self.index.shape == data.index.shape: # index may be different if schedule is udpated
    #             self.index = data.index
    #             self.loc(push=True, update=True)[:, data.columns] = data.values
    #         else:
    #             self.loc(push=True, update=True)[data.index, data.columns] = data.values
    #     elif isinstance(data, pd.Series):
    #         self.loc(push=True, update=True)[data.index, :] = data
    #     else:
    #         self.iloc(push=True, update=True)[:len(data)] = data

    def to_frame(self, with_periods:bool=True, strftime:str=None):
        frame = super().to_frame()
        if not with_periods and isinstance(frame.columns, pd.PeriodIndex):
            strftime = PERIODS_TO_STRF[frame.columns.freqstr] if strftime is None else strftime
            frame.columns = frame.columns.strftime(strftime)
        return frame

    def to_account(self):
        if self.shape[0] > 1:
            raise ValueError('Only MultiLineItem objects with a single row can be converted to LineItem objects.')
        
        return LineItem(
            self.values[0],  
            index=self.columns,
            **self._metaparams
        )



import numpy as np
import pandas as pd

from .base import statfunc

__all__ = ['add', 'multiply', 'divide', 'cumsum', 'compound_growth']

def derag(args):
    if any(not isinstance(arg, (pd.Series, pd.DataFrame, np.ndarray)) for arg in args):
        args = np.array(args, dtype='object')
    return args

@statfunc
def add(*args):
    return np.add.reduce(derag(args))

@statfunc
def multiply(*args):
    return np.multiply.reduce(derag(args))

@statfunc
def divide(a, b):
    a, b = derag((a, b))
    if isinstance(a, (pd.Series, np.ndarray)):
        a = a.astype('float')
    if isinstance(b, (pd.Series, np.ndarray)):
        b = b.astype('float')

    return np.divide(a, b, out=np.zeros_like(a), where=b!=0)

@statfunc
def cumsum(arr):
    return np.cumsum(arr)

@statfunc
def compound_growth(init, g, n):
    return init * np.repeat(1 + g, n).cumprod()

@statfunc
def repeat(value, n, start=0):
    arr = np.zeros(n)
    arr[start:] = np.repeat(value, n - start)
    return arr


import functools as ft
import warnings
import typing as typ
import numpy as np
import pandas as pd

import networkx as nx

from .base import FinStatBaseObject, PERIODS_TO_STRF
from .elements import LineItem, MultiLineItem, Elements, Schedule, ScheduleFunction, ScheduleFactory

from .functions.base import StatFunction
from .optimize import ModelManager
from .groupers import FinStatResampler, FinStatGroupBy
from .valuation import Valuation

NODE_TYPES = np.array(['factor', 'time-series-factors', 'statement', 'schedule', 'account', 'metric'])
EXCLUDED_NODES = NODE_TYPES[:-2]
LINEITEM_NODES = NODE_TYPES[-2:]
NON_FACTOR_NODES = NODE_TYPES[1:]
REQUIRED_NODE_ATTRIBUTES = [
    'obj', 'name', 'short_name', 'statement', 
    'nodetype', 'is_input', 'is_multi', 'is_total', 'model', 'expr', 
    'position', 'insert_after', 'hide'
    ]

FinancialStatementType = typ.TypeVar('FinancialStatementType', bound='FinancialStatement')

class ModelContextManager:
    def __init__(self, stat, target:str=None, solve:bool=True, new_instance:bool=False, *args, **kwargs):
        self.stat = stat
        self._solve = solve
        self._new_instance = new_instance

        self._tgtstr = target

    @property
    def M(self):
        return self.stat.M

    def assign_objective(self, target):
        self._tgtstr = target
        self.M.assign_objective(self._tgtstr)

    # Utility functions to update graph attributes
    def update_node_expr(self, expr, short_name):
        self.stat.G.nodes[short_name]['expr'] = expr

    def add_nodeholder(self, 
        name, 
        short_name, 
        model,
        is_multi=False,
        is_total=False,
        hide=False, 
        insert_after=None,
        ):
        """
        Creates model attributes, constraints, and graph node associated with a new metric.

        Both the expression and the finalized object can be appended at any point
        """
        self.stat.G.add_node(
            short_name,
            obj=None,
            name=name,
            short_name=short_name,
            statement=self.stat.short_name,
            model=model,
            expr=None,
            nodetype='metric',
            is_input=False,
            is_multi=is_multi,
            is_total=is_total,
            hide=hide,
            position=self.stat.G.find_node_position(insert_after),
            insert_after=insert_after,
        )

    # User-exposed full operations #
    def as_periods(self, expr=None, construct:bool=True, assign:bool=True, short_name=None, is_multi=False, **kwargs):
        # turns the single ScalarVar expression into a list of IndexedVar expressions for each period
        # and adds each expression to corresponding constraint
        # expression added to the existing node;
        # this is the expr BEFORE the periodize transformation
        if isinstance(expr, str) and 'name' not in kwargs:
            raise ValueError('If you do not provide `expr` because `assignt=False` you must provide `name` as kwarg')

        if not construct and short_name is None:
            raise ValueError('When `construct=False`, you must provide `short_name` in order to identify the already constructed node.')

        if short_name is None and 'name' in kwargs:
            short_name = self.stat._shorten(kwargs['name'])

        if construct and not self.stat.has_node(short_name):
            if not self.stat.no_model:
                self.M.assign_vars(short_name, is_multi=is_multi) # instantiate model relations (root Var, IndexedVar, Constraint List)
                model = self.M.get(short_name)
            else:
                model = 1 # when it is called, allows model attribute to be added to another
            self.add_nodeholder(short_name=short_name, is_multi=is_multi, model=model, **kwargs) # add corresponding node (with expr=None)

        if assign:
            assert expr is not None
            if not self.stat.no_model:
                self.M.periodize(short_name, expr, is_multi=is_multi) # periodize model expressions and assign constraints
                self.update_node_expr(expr, short_name) # update node for the regular expression on the root Vars / Params

    def as_multi(self, *args, **kwargs):
        self.as_periods(*args, is_multi=True, **kwargs)

    def as_total(self, comp, construct:bool=True, assign:bool=True, short_name=None, **kwargs):
        # if isinstance(expr, str) and 'name' not in kwargs:
        #     raise ValueError('If you do not provide `expr` because `assign=False` you must provide `name` as kwarg')

        if not construct and short_name is None:
            raise ValueError('When `construct=False`, you must provide `short_name` in order to identify the already constructed node.')

        if short_name is None:
            short_name = f'tot_{comp.name}'

        if construct and not self.stat.has_node(short_name):
            self.M.assign_vars(short_name, is_multi=False) # instantiate model relations (root Var, IndexedVar, Constraint List)
            self.add_nodeholder(short_name=short_name, is_multi=False, is_total=True, **kwargs) # add corresponding node (with expr=None)

        if assign:
            self.M.total(comp) # periodize model expressions and assign constraints
            self.update_node_expr(comp, short_name) # update node for the regular expression on the root Vars / Params

    def where(self, expr=None, big_M=1000, construct:bool=True, assign:bool=True, short_name=None, **kwargs):
        if isinstance(expr, str) and 'name' not in kwargs:
            raise ValueError('If you do not provide `expr` because `assignt=False` you must provide `name` as kwarg')

        if not construct and short_name is None:
            raise ValueError('When `construct=False`, you must provide `short_name` in order to identify the already constructed node.')

        if short_name is None and 'name' in kwargs:
            short_name = self.stat._shorten(kwargs['name'])

        where = self.M.where(short_name, big_M=big_M) # instantiate constructor only
         
        if construct and not self.stat.has_node(short_name):
            where.assign_vars()

            self.add_nodeholder(**kwargs) # actual node

            kwargs['name'] = 'delta_' + kwargs['short_name']
            kwargs['short_name'] = kwargs['name']
            self.add_nodeholder(**kwargs) # delta node

        if assign:
            where.parse_inequal(expr)
            where.assign_exprs()
            self.update_node_expr(expr, short_name)

    def __call__(self, *args, **kwargs):
        self.as_periods(*args, **kwargs)

    def __enter__(self):
        self.stat.open_model_context() # Open model context so that the statement attributes return Pyomo expressions
        return self
    
    def __exit__(self, exc_type, exc_value, exc_tb):
        try:
            if exc_type:
                raise exc_type(exc_value) # preempt other operations by raising any incoming exception
            
            if not self.stat.no_model:
                if not self.M.obj_assigned:
                    self.M.assign_objective(self._tgtstr)
                
                if self._new_instance or not self.M.has_instance:
                    self.M.create_instance()

                if self._solve:
                    self.M.solve()
        finally:
            self.stat.close_model_context() # Flag must be closed after calculation to ensure LineItem objects can be used normally

class StatGraph:
    REQUIRED_ATTRS = REQUIRED_NODE_ATTRIBUTES
    NODE_TYPES = NODE_TYPES

    def __init__(self, stat:FinancialStatementType, graph:nx.Graph):
        self._stat = stat
        self._graph = graph # each statement has its own StatGraph, but all sub-statements shared the same nx Graph

    def __getattribute__(self, name:str) -> typ.Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError as e:
            try:
                graph = object.__getattribute__(self, '_graph')
                return object.__getattribute__(graph, name)
            except AttributeError:
                pass

            raise e

    def add_node(self, node:str, **kwargs):
        """
        Centralizes `add_node` to ensure each node receives the same attributes.
        """
        not_provided = [attr for attr in self.REQUIRED_ATTRS if attr not in kwargs]
        assert not not_provided, f'`add_node` has missing parameters: {", ".join(not_provided)}'
        extra_provided = [k for k in kwargs.keys() if k not in self.REQUIRED_ATTRS]
        assert not extra_provided, f'`add_node` given extra parameters: {", ".join(extra_provided)}'

        self._graph.add_node(node, **kwargs)

    ### GRAPH FILTERING ###
    def filter_subgraph(self, attr, value, statname):
        # returns a subgraph containing only the nodes with attribute value specified
        graph = self._graph
        def filter_stat(n):
            return graph.nodes(data='statement')[n] == statname
        
        def f(n):
            if hasattr(value, '__len__') and not isinstance(value, str):
                return graph.nodes(data=attr)[n] in value
            else:
                return graph.nodes(data=attr)[n] == value

        if statname in graph.nodes: # if the statement caller is a node, then it is a sub-statement and we want to filter the graph to just that 
            statgraph = nx.subgraph_view(graph, filter_node=filter_stat)
            return nx.subgraph_view(statgraph, filter_node=f)
        else:
            return nx.subgraph_view(graph, filter_node=f)
    
    def filter_graph_by_attribute(self, attr, value):
        return self.filter_subgraph(attr, value, self._stat.short_name)

    def filter_graph_by_attributes(self, **kwargs):
        for k, v in kwargs.items():
            graph = self.filter_graph_by_attribute(k, v)
        return graph

    def filter_nodes_by_attribute(self, attr, value, return_node:bool=False, return_tuple:bool=False, **kwargs):
        for node in self.filter_subgraph(attr, value, self._stat.short_name).nodes(**kwargs):
            if return_node:
                yield self._graph.nodes[node]
            else:
                node = node[1] if isinstance(node, tuple) and not return_tuple else node
                if 'data' in kwargs and kwargs['data'] == 'obj':
                    if isinstance(node, tuple):
                        yield node[0], node[1]() if callable(node[1]) else node[1] # calls the ScheduleFunction
                    else:
                        yield node() if callable(node) else node
                else:
                    yield node

    def filter_nodes_by_attributes(self, return_tuple=False, **kwargs):
        filter_kwargs = {k: v for k, v in kwargs.items() if k in self.REQUIRED_ATTRS}
        kwargs = {k: v for k, v in kwargs.items() if k not in filter_kwargs}
        graph = self.filter_graph_by_attributes(**filter_kwargs)

        for node in graph.nodes(**kwargs):
            node = node[1] if isinstance(node, tuple) and not return_tuple else node
            if 'data' in kwargs and kwargs['data'] == 'obj':
                yield node() if callable(node) else node # calls the ScheduleFunction
            else:
                yield node

    ### POSITION MANAGEMENT ###
    @property
    def node_positions(self):
        return np.array(list(nx.get_node_attributes(self._graph, 'position').values()))

    @property
    def next_node_position(self):
        if self.node_positions.size == 0:
            return 0
        else:
            return self.node_positions.max() + 1

    def get_node_position(self, name):
        """
        Returns order attribute of node
        """
        return self._graph.nodes[name]['position']

    def update_node_positions(self, insert_pos:int):
        old_order = self.node_positions
        new_order = np.where(old_order >= insert_pos, old_order + 1, old_order)
        for name, position in zip(list(self._graph.nodes), new_order):
            self._graph.nodes[name]['position'] = position

    def find_node_position(self, insert_after:typ.Union[str, None]):
        if insert_after is None:
            order = self.next_node_position
        else:
            order = self.get_node_position(insert_after) + 1
            self.update_node_positions(order)

        return order

class ViewManager:
    SUPPORTED_VIEWS = ['visible', 'hidden', 'all']
    NODE_TYPES = NODE_TYPES
    EXCLUDED_NODES = EXCLUDED_NODES
    LINEITEM_NODES = LINEITEM_NODES

    def __init__(self, stat:FinancialStatementType):
        self._stat = stat
        self._viewstr = 'visible'

    def set_view(self, viewstr:str):
        if self._viewstr not in self.SUPPORTED_VIEWS:
            raise ValueError(f"`viewstr` must be one of {', '.join(self.SUPPORTED_VIEWS)}")

        self._viewstr = viewstr

    def visible(self):
        self.set_view('visible')
        return self._stat

    def all(self):
        self.set_view('all')
        return self._stat

    def get_visible(self):
        filters = dict(nodetype=ViewManager.LINEITEM_NODES, statement=self._stat.short_name, hide=False)
        return Elements(*self._stat.G.filter_nodes_by_attributes(data='obj', **filters))

    def get_all(self):
        filters = dict(nodetype=self.LINEITEM_NODES, statement=self._stat.short_name)
        return Elements(*self._stat.G.filter_nodes_by_attributes(data='obj', **filters))

    def __call__(self):
        """
        Attribute manages the contents of `to_frame` output
        """
        return getattr(self, f'get_{self._viewstr}')()

class FinancialStatement(FinStatBaseObject):
    """
    Main class
    
    Consists of DataFrame front-end, matrix backend?,
    and graph structure to map front-end on to backend,
    graph structure used to facilitate observer pattern
    and automatic updates

    Financial statement as a directed graph
    Each LineItem / Metric / Ratio is a node
    LineItem objects never have predecessors
    Metric objects must have predecssors

    Each account is assigned to a 
            self.graph.add_node(account.short_name, obj=account)

    Parameters:
        periods: pd.PeriodIndex
            time-series applied to all the constituent lineitems
    """
    _metadata = ['_name', '_short_name', '_graph']
    EXCLUDED_NODES = EXCLUDED_NODES
    LINEITEM_NODES = LINEITEM_NODES
    _ipython_canary_method_should_not_exist_ = None # Need for ipython support of __getattribute__ customization

    def __init__(self, 
        periods:pd.PeriodIndex=None, 
        name:str=None,
        short_name:str=None,
        groupby:typ.Iterable[str] | str=[],
        graph:nx.DiGraph=None,
        no_model:bool=False,
        ):
        if graph is None:
            graph = nx.DiGraph(
                stat=self,
                OPEN_MODEL_CONTEXT=False,
                periods=periods,
                no_model=no_model,
                by=[groupby] if isinstance(groupby, str) else groupby # this can be access by any object with the graph as an attribute
            )
        
        super().__setattr__('_graph', StatGraph(self, graph))
        super().__setattr__('_name', name)
        super().__setattr__('_short_name', self._shorten(name, short_name))
        super().__setattr__('_view_mngr', ViewManager(self))

        if 'model_mngr' not in self._graph.graph:
            self._graph.graph['model_mngr'] = ModelManager(self)

        super().__setattr__('RELATED', [])

    @property
    def _constructor(self):
        return FinancialStatement

    def __getattribute__(self, name:str) -> typ.Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError as e:
            try:
                graph = object.__getattribute__(self, 'G')
                nodes = graph.__getattribute__('nodes')
                if name in nodes:
                    AS_EXPR = object.__getattribute__(self, 'AS_EXPR') ### when AS_EXPR, the pyomo model version of the node attribute is passed
                    if AS_EXPR:
                        return nodes[name]['model']
                    else:
                        obj = nodes[name]['obj']
                        if callable(obj): # if 'obj' is a ScheduleFunction, call the it
                            return obj()
                        else:
                            return obj
            except AttributeError as e:
                raise e

            raise e

    def __setattr__(self, key, value):
        if self.has_node(key):
            node = self.graph.nodes[key]
            if node['nodetype'] == 'factor':
                M_attr = self.M.get(key)
                M_attr = value # assigning the value directly will update the already-initialized Pyomo object (should be a ScalarParam)
                self.solve_out_of_context() # resolve model if there is an objective function
                node['obj'] = value
        elif isinstance(value, (int, float, np.number)):
            self.add_factor(key, value)
        else:
            raise Exception('FinancialStatement objects do not allow attribute assignment except for `factor` nodes')

    def get(self, name):
        if self.model_context_is_open:
            warnings.warn('When `OPEN_MODEL_CONTEXT=True`, `get` returns a pyomo object, not a finstat object')
        return getattr(self, name)

    ### VISUALIZAITON ###
    def __repr__(self):
        return f'FinancialStatement: {self.short_name}'

    def _repr_html_(self):
            if self.is_substat or self.statements.size == 0:
                return self._frame._repr_html_()
            else:
                statstrs = ''
                last = 0
                for i, (short_name, name) in enumerate(zip(self.statements.short_names, self.statements.names)):
                    statstrs += f"""
                        <tr style="text-align: left;">
                        <td>{i + 1}.</td>
                        <td style="text-align: left;">{short_name}</td>
                        <td style="text-align: left;">{name}</td>
                        </tr>
                    """
                    last = i
                for i, related in enumerate(self.RELATED):
                    statstrs += f"""
                        <tr style="text-align: left;">
                        <td>{last + i + 1}.</td>
                        <td style="text-align: left;">{related.short_name}</td>
                        <td style="text-align: left;">{related.name}</td>
                        </tr>
                    """
                return f"""<div>\n
                    <table border="1" class="dataframe">\n  
                    <caption style="text-align: left; font-size: 1.25em; font-weight: bolder;">Statements</caption>
                    <thead>\n    
                    <tr style="text-align: left;">\n
                        <th></th>\n
                        <th style="text-align: left;">Attr</th>\n    
                        <th style="text-align: left;">Name</th>\n
                    </tr>\n
                    </thead>\n  
                    <tbody>\n
                    {statstrs}
                    </tbody>\n
                    </table>\n
                </div>
                """
    
    @property
    def no_model(self):
        return self.G.graph['no_model']

    @property
    def view(self):
        """
        Attribute manages the contents of `to_frame` output
        """
        return self._view_mngr

    @property
    def _frame(self):
        super().__setattr__('__frame__', self.to_frame())
        return self.__frame__

    def to_frame(self, with_periods:bool=True, strftime:str=None, *args, **kwargs) -> pd.DataFrame:
        if self.empty:
            frame = pd.DataFrame([], columns=[f'Empty Statement: {self.name}'])
        elif self.view().size == 0:
            frame = pd.DataFrame([], columns=[f'Empty View: {self.name}'])
        else:
            frame = self.view().concat(*args, by=self.by, **kwargs)

        if not with_periods and isinstance(frame.columns, pd.PeriodIndex):
            if strftime is None:
                strftime = PERIODS_TO_STRF[self.periods.freqstr]
            
            frame.columns = frame.columns.strftime(strftime)

        return frame

    #### Standard Properties ####
    @property
    def name(self):
        return self._name

    @property
    def _periods(self):
        return self.graph.graph['periods']

    @property
    def periods(self):
        return self._periods

    def update_periods(self, periods:pd.PeriodIndex):
        self.graph.graph['periods'] = periods 

    @property
    def empty(self):
        return not self.lineitems.size

    @property
    def by(self):
        return self.graph.graph['by']

    def update_by(self, by:typ.Iterable[str]):
        self.graph.graph['by'] = by

    @property
    def is_substat(self):
        return self.short_name in self.G.nodes # the main statement object will not be a node in the graph; all others will be

    ### Nodes ###
    @property
    def elements(self):
        graph = object.__getattribute__(self, 'G')
        return Elements(*graph.filter_nodes_by_attribute('nodetype', NON_FACTOR_NODES, data='obj'))

    @property
    def statements(self):
        graph = object.__getattribute__(self, 'G')
        return Elements(*graph.filter_nodes_by_attribute('nodetype', 'statement', data='obj'))    

    @property
    def factors(self):
        filters = dict(nodetype='factor')
        if self.is_substat:
            filters['statement'] = self.short_name

        factors = self.G.filter_nodes_by_attributes(data='obj', return_tuple=True, **filters)
        return pd.Series(dict(factors), name='Factors', dtype=object)

    @property
    def schedules(self):
        filters = dict(nodetype='schedule', statement=self.short_name)
        return Elements(*self.G.filter_nodes_by_attributes(data='obj', **filters))

    @property
    def lineitems(self) -> Elements:
        """
        Returns all LineItem objects that are nodes on the FinancialStatement graph. This includes lineitems nested
        in the subgraphs of sections.
        """
        filters = dict(nodetype=self.LINEITEM_NODES)
        return Elements(*self.G.filter_nodes_by_attributes(data='obj', **filters))

    @property
    def accounts(self) -> Elements:
        """
        Returns all LineItem objects that are nodes on the FinancialStatement graph. This includes lineitems nested
        in the subgraphs of sections.
        """
        return Elements(*self.G.filter_nodes_by_attributes(data='obj', nodetype='account'))

    def add_related(self, name, statement):
        # RELATED ARE NOT A PART OF THE GRAPH!!!
        super().__setattr__(name, statement)
        self.RELATED.append(statement)

    def add_statement(self, name, periods=None, insert_after:str=None, **kwargs):
        """
        Adds statement meta to the FinancialStatement
        1. Add statement meta to graph
        2. Add statement `short_name` to `__statements__` (need for lookup in `__getattribute__`)
        3. Assign `short_name` as attribute that simply returns the statement object
        """
        periods = self.periods if periods is None else periods
        stat = FinancialStatement(name=name, graph=self.graph._graph, periods=periods, **kwargs) # pass the underlying nx Graph, NOT the StatGraph

        self.G.add_node(
            stat.short_name,
            obj=stat,
            name=stat.name,
            short_name=stat.short_name,
            statement=self.short_name,
            nodetype='statement',
            is_input=False,
            is_multi=False,
            is_total=False,
            expr=None,
            model=None,
            position=self.G.find_node_position(insert_after),
            insert_after=insert_after,
            hide=None,
        )

    def add_factor(self, name:str, value:typ.Union[np.number, int, float, str, pd.DataFrame, np.ndarray, list]=None):
        if ' ' in name:
            name = self._shorten(name)

        if not self.no_model and isinstance(value, (int, float)):
            self.M.setparam(name) # only Real numbers can be added as params to model
            model = self.M.get(name)
        else:
            model = None

        self.G.add_node(
            name,
            obj=value,
            name=name,
            short_name=name,
            statement=self.short_name,
            nodetype='factor',
            is_input=True,
            is_multi=False,
            is_total=False,
            expr=None,
            model=model,
            position=self.G.find_node_position(None),
            insert_after=None,
            hide=None,
        )

    def add_factors(self, factors:typ.Iterable):
        if isinstance(factors, dict):
            for k,v in factors.items():
                self.add_factor(k, v)
        else:
            for factor in factors:
                self.add_factor(**factor)

    def add_schedule(self, 
        schedule:typ.Union[Schedule, typ.Callable], 
        name:str='',
        short_name:str=None,
        insert_after:str=None,
        **kwargs
        ):
        if callable(schedule):
            schedule = ScheduleFactory(schedule, self)
            schedule.set_call_kwargs(name=name, short_name=short_name, graph=self.graph, **kwargs)
        else:
            schedule.set_graph(self.graph)
    
        self.G.add_node(
            schedule.short_name,
            obj=schedule,
            name=schedule.name, 
            short_name=schedule.short_name,
            statement=self.short_name,
            nodetype='schedule',
            is_input=False,
            is_multi=False,
            is_total=False,
            model=None,
            expr=None,
            position=self.G.find_node_position(insert_after),
            insert_after=insert_after,
            hide=True,
        )

    def add_schedules(self, *schedules:typ.Iterable[Schedule], common_kws={}):
        for sched in schedules:
            self.add_schedule(sched, **common_kws)

    def add_account(
        self, 
        data:typ.Iterable,
        name:str='',
        short_name:str=None,
        periods:pd.PeriodIndex=None,
        index:typ.Iterable=None,
        insert_after:str=None,
        hide:bool=False,
        **kwargs
        ) -> None:
        """
        An lineitem is a pandas object with no predecessors. When an lineitem is added,
        1. it is added as an attribute according to its `short_name`
        2. the Graph is passed into the LineItem
        3. the LineItem is added as a node in the graph

        Parameters:
            data: any python iterables
            name: str
            short_name: str
                used to assign attribute name to the FinancialStatement
            periods: pd.PeriodIndex
                can be passed directly to override class-level attribute
        """
        is_input = True # only time it is not is if it is a function!
        if isinstance(data, (LineItem, MultiLineItem)):
            lineitem = data
            lineitem._graph = self.graph
            if not lineitem.periods.equals(self.periods):
                lineitem.update_periods(self.periods)
        elif isinstance(data, (StatFunction, ScheduleFunction, ScheduleFactory)):
            data.set_call_kwargs(name=name, short_name=short_name, graph=self.graph, **kwargs)
            lineitem = data
            is_input = False
        else:
            if isinstance(data, pd.Series):
                if not name:
                    name = data.name
                data, index = data.values, data.index

            if not name:
                raise ValueError('You must provide `name` attribute if `data` is not type pd.Series')

            data = np.array(data) # To check dimensionality
            periods = periods if periods is not None else self.periods
            if data.ndim == 2:
                lineitem = MultiLineItem(
                    data, 
                    index=index,
                    columns=periods,
                    name=name, 
                    short_name=short_name, 
                    graph=self.graph,
                    **kwargs
                )
            else:
                lineitem = LineItem(
                    data, 
                    index=periods, 
                    name=name, 
                    short_name=short_name, 
                    graph=self.graph,
                    **kwargs
                )

        if lineitem.is_multi:
            self.M.set_multi_index()
        
        if not self.no_model:
            self.M.setparam(lineitem.short_name) # root attribute is a shadow used to generate expressions; must be Var to allow for expressions
            self.M.setparam( f'{lineitem.short_name}_periods', is_multi=lineitem.is_multi) # assign the IndexedParam; this contains actual data and is used in expression transformation
            model = self.M.get(lineitem.short_name)  # the node contains the uninitialized root Var in order to generate epxressions
        else:
            model = 1 # use string so that value can be added/subtracted

        self.G.add_node(
            lineitem.short_name, 
            obj=lineitem,
            name=lineitem.short_name,
            short_name=lineitem.short_name,
            expr=None,
            model=model,
            statement=self.short_name,
            is_input=is_input,
            is_multi=lineitem.is_multi,
            is_total=False,
            nodetype='account',
            position=self.G.find_node_position(insert_after),
            insert_after=insert_after,
            hide=hide
        )
       
    def add_accounts(self, data:typ.Union[pd.DataFrame, LineItem, MultiLineItem, typ.Iterable], *args, common_kws={}) -> None:
        """
        Add multiple accounts with one call.

        If data is a pd.DataFrame, can iterate over the pd.Series rows. Otherwise iterate over the collections.
        """
        if isinstance(data, pd.DataFrame) and not isinstance(data, MultiLineItem):
            for i, row in data.iterrows():
                self.add_account(data=row, **common_kws)
        else:
            if not isinstance(data, (list, tuple, np.ndarray)):
                data = [data]
                if len(args) > 0:
                    data += args

            for account in data:
                if isinstance(account, (LineItem, MultiLineItem)):
                    self.add_account(account, **common_kws)          
                else:
                    self.add_account(**account, **common_kws)

    ### Attached Modules ###
    def add_metrics(self, *args, **kwargs):
        return ModelContextManager(self, *args, **kwargs)

    def valuation(self, *args, **kwargs):
        return Valuation(*args, **kwargs)

    ##### Implementations of common pandas methods #####
    @property
    def iloc(self):
        return self._frame.iloc

    @property
    def loc(self):
        return self._frame.loc

    @property
    def index(self):
        return self._frame.index

    @property
    def shape(self):
        return self._frame.shape

    def resample(self, *args, **kwargs):
        return FinStatResampler(self, *args, **kwargs)

    def groupby(self, *args, **kwargs):
        return FinStatGroupBy(self, *args, **kwargs)

    def reorder_levels(self, *args, **kwargs):
        warnings.warn('`reorder_levels` currently returns a `MultiLineItem` object, NOT a FinancialStatement')
        return self._frame.reorder_levels(*args, **kwargs).sort_index()

    def droplevel(self, *args, **kwargs):
        warnings.warn('`lineitems_only` currently returns a `MultiLineItem` object, NOT a FinancialStatement')
        return self._frame.droplevel(*args, **kwargs)

    def to_csv(self, *args, **kwargs):
        self._frame.to_csv(*args, **kwargs)



import typing as typ
import numpy as np
import pandas as pd

def cagr(fv, pv, t):
    """
    Computes compound annual growth rate implied by two values over a time, t

    Params
    -------
    fv:     float, future value
    pv:     float, present value
    t:      time between values in years
    """
    return (fv / pv)**(1/t) - 1

def discount_factors(cf, discount, t):
    return (1 + discount)**np.arange(t, cf.shape[0] + t)

def wacc_calc(cost_debt, cost_equity, debt_to_cap):
    return (cost_debt*debt_to_cap) + (cost_equity*(1-debt_to_cap))

def npv(cf, disc=.08):
    """
    Net Present Value calculation for a stream of cash flows
    
    Params
    -------
    cf:    nxm array of cash flow streams
            > n = # of years
            > m = # of different cash flow streams
    disc:  float representing annual discount rate
            > default value per Duff&Phelps recommendation here:
                https://www.duffandphelps.com/-/media/assets/pdfs/publications/articles/dp-erp-rf-table-2020.pd
    
    Returns
    --------
        nxm array representing of each cash inflow discounted to the current day
    """
    return cf.values.reshape(-1,cf.ndim) / (1+disc)**np.arange(0, cf.shape[0]).reshape(-1,1)

def logparams(x1, y1, x2, y2):
    """
    Returns parameters for line of form 
    
    y = a*ln(b*x)
    
    https://math.stackexchange.com/questions/716152/graphing-given-two-points-on-a-graph-find-the-loglarithmic-function-that-passes
    """
    a = (y1 - y2) / np.log(x1 / x2)
    b = np.exp((y2*np.log(x1) - y1*np.log(x2)) / (y1 - y2))
    
    return a, b

def logline(x, a, b):
    xdim = b.shape[0]
    x = x.repeat(xdim).reshape(-1,xdim)
    a = a.reshape(-1,1)
    b = b.reshape(-1,1)

    return a*np.log(b*x.T)

def log_returns(ser):
    if not isinstance(ser, pd.Series):
        ser = pd.Series(ser)
    return np.log(ser / ser.shift(1)).iloc[1:]

class Valuation:
    def __init__(self, cash_flows:typ.Iterable, discount:float, fwd_multiple:float, t=1):
        self.cash_flows = cash_flows
        self.discount = discount
        self.fwd_multiple = fwd_multiple
        self.t = t
        
        idx = self.cash_flows.index
        self.tvidx = pd.period_range(start=idx[0], periods=idx.size + 1, freq=idx.freqstr)
    
    @property
    def discount_factors(self):
        return discount_factors(self.cash_flows_with_tv(), self.discount, self.t)
    
    @property
    def terminal_value(self):
        return self.cash_flows.iloc[-1] * self.fwd_multiple        

    @property
    
    def tv(self):
        return self.terminal_value
    
    @property
    def discounted_tv(self):
        return self.discounted_cash_flows().iloc[-1]

    def cash_flows_with_tv(self):
        ser = pd.concat((self.cash_flows, pd.Series([self.terminal_value])))
        ser.index = self.tvidx
        return ser
    
    def discounted_cash_flows(self):
        return self.cash_flows_with_tv() / self.discount_factors

    def fair_value(self):
        return self.discounted_cash_flows().sum()
    
    def tv_per_fv(self):
        return self.discounted_tv / self.fair_value()
    
    def __call__(self):
        return self.fair_value()



import inspect
import functools as ft

from ..base import FinStatBaseObject
from ..elements.lineitem import LineItem

class statfunc:
    def __init__(self, func):
        ft.update_wrapper(self, func)
        self.func = func

    def __call__(self, *args, **kwargs):
        return ArrayFunction(self.func, *args, **kwargs)

class FuncMixin:
    _call_kwargs = {}

    def _shorten(self, *args, **kwargs):
        return FinStatBaseObject._shorten(self, *args, **kwargs)

    @property
    def call_kwargs(self):
        return self._call_kwargs

    def set_call_kwargs(self, **kwargs):
        if not kwargs['short_name']:
            kwargs['short_name'] = self._shorten(kwargs['name'])
        
        self._call_kwargs = kwargs
        
        for k, v in self.call_kwargs.items():
            setattr(self, k, v)

class StatFunction(FuncMixin):
    pass

class ArrayFunction(StatFunction):
    def __init__(self, func, *fargs, **fkwargs):
        ft.update_wrapper(self, func)

        funcsig = inspect.getfullargspec(func)
        if 'periods' in fkwargs and 'periods' in funcsig.args:
            self._periods = fkwargs['periods']
        elif 'periods' in inspect.getargspec(func).args:
            self._periods = fargs[funcsig.args.index('periods')]
        else:
            self._periods = fkwargs.pop('periods', None)

        self.func = func
        self.fargs = fargs
        self.fkwargs = fkwargs

    def __repr__(self):
        return f'Array Function: {self.__name__}'

    @property
    def is_multi(self):
        return False # allows StatFunction to act as a LineItem in `add_account` method

    @property
    def periods(self):
        if hasattr(self, '_periods') and self._periods is not None:
            return self._periods
        else:
            args = list(self.fargs) + list(self.fkwargs.values())
            for arg in args:
                if hasattr(arg, 'has_graph'):
                    if arg.has_graph:
                        return self.graph.graph['periods']

            raise ValueError('You must provide `periods` argument if none of the function arguments are part of a FinancialStatement')

    def __call__(self, *args, **kwargs):
        obj = self.func(*args, *self.fargs, **kwargs, **self.fkwargs)
        return LineItem(
            obj,
            index=self.periods,
            *args,
            **kwargs,
            **self.call_kwargs
        )



import functools as ft
import typing as typ
import numpy as np
import pandas as pd

import pyomo.environ as pyo
from pyomo.core.expr.numvalue import nonpyomo_leaf_types
from pyomo.core.expr import current as EXPR
from pyomo.core.base.var import IndexedVar
from pyomo.opt import TerminationCondition
from pyomo.core.expr.logical_expr import InequalityExpression
from pyomo.core.expr.current import ExpressionBase
from pyomo.core.base.param import IndexedParam, ScalarParam

from .elements import LineItem, MultiLineItem
from .base import PERIODS_TO_STRF

"""
Custom Expressions
--------------------------------------
"""
def model_finder(expr):
    if hasattr(expr, 'parent_block'):
        model = expr.parent_block()
    else:
        if hasattr(expr, 'args'):
            for item in expr.args:
                if  hasattr(item, 'parent_block'):
                    model = item.parent_block()
                    break
    
    return model

def prior(expr, *args, **kwargs):
    model = model_finder(expr)
    return PriorExpression(expr, model, *args, **kwargs)

def cumsum(expr, *args, **kwargs):
    model = model_finder(expr)
    return CumSumExpression(expr, model, *args, **kwargs)

def mean(expr, *args, **kwargs):
    model = model_finder(expr)
    return MeanExpression(expr, model, *args, **kwargs)

class MethodExpression(ExpressionBase):
    """
    https://pyomo.readthedocs.io/en/latest/developer_reference/expressions/managing.html#walking-an-expression-tree-with-a-visitor-class
    """
    PRECEDENCE = 0

    def __init__(self, expr, model, index=None):
        super().__init__((expr,))
 
        self._root_expr = expr
        self._model = model
        self._index = index # required only for testing; arg passed to "create_node" should have the period from `periodize`

    def set_index(self, index):
        self._index = index

    def set_model(self, model):
        self._model = model # backdoor to update expression during model construction ... without, certain parameters are not initialized by the time expressino is evauluated

    def nargs(self):
        return 1 # target node is the only argument

    def _compute_polynomial_degree(self, result):
        return result[0]
    
    def _apply_operation(self, result):
        return result

class CumSumExpression(MethodExpression):    
    def create_node_with_local_data(self, args, classtype=None):
        """
        Returns cumulative summation of the indexed variable up to the period specified.
        """
        assert classtype is None
        arg = args[0]
        if arg is self._root_expr: # this SHOULD only be used when testing; args passed should be same as __init__
            indexed = getattr(self._model, f'{arg.name}_periods')
        else:
            indexed = getattr(self._model, arg.name.split("[")[0]) # here, arg is single value so must find full IndexedVariable from whence it came
        
        last = self._model.periods.ord(self._index) + 1
        node = sum(indexed[self._model.periods.at(i)] for i in range(1, last))
        
        return node

def compound_growth(base, growth_rate, n):
    #### SHOULD BE AN EXPRESSION!!!!!####
    factors = np.repeat(1 + growth_rate, n).cumprod()
    return base*factors

class CompoundGrowthExpression(MethodExpression):
    def create_node_with_local_data(self, args, classtype=None):
        """
        Returns cumulative summation of the indexed variable up to the period specified.
        """
        assert classtype is None

        arg = args[0]
        if arg is self._root_expr: # this SHOULD only be used when testing; args passed should be same as __init__
            indexed = getattr(self._model, f'{arg.name}_periods')
        else:
            indexed = getattr(self._model, arg.name.split("[")[0]) # here, arg is single value so must find full IndexedVariable from whence it came
        
        last = self._model.periods.ord(self._index) + 1
        node = sum(indexed[self._model.periods.at(i)] for i in range(1, last))
        
        return node

class PriorExpression(MethodExpression):
    """
    https://pyomo.readthedocs.io/en/latest/developer_reference/expressions/managing.html#walking-an-expression-tree-with-a-visitor-class
    """
    PRECEDENCE = 20

    def __init__(self, expr, model, short_name, index=None):
        super().__init__(expr, model, index=index)
 
        self._short_name = short_name
    
    def create_node_with_local_data(self, args, classtype=None):
        """
        Returns cumulative summation of the indexed variable up to the period specified.
        """
        assert classtype is None

        node = args[0]

        iloc = self._model.periods.ord(self._index)
        indexed = getattr(self._model, f'{self._short_name}_periods')
        if iloc != 1: # if the period is the first period, there is no prior period
            node = node + indexed[self._model.periods.at(iloc - 1)] # for every period other than the first period, get the prior period
                    
        return node

class MeanExpression(MethodExpression):
    """
    https://pyomo.readthedocs.io/en/latest/developer_reference/expressions/managing.html#walking-an-expression-tree-with-a-visitor-class
    """
    PRECEDENCE = 20

    def mean(self, t1, t2):
        return (t1 + t2) / 2

    def create_node_with_local_data(self, args, classtype=None):
        """
        Returns cumulative summation of the indexed variable up to the period specified.
        """
        assert classtype is None

        arg = args[0]
        if arg is self._root_arg: # this SHOULD only be used when testing; args passed should be same as __init__
            indexed = getattr(self._model, f'{arg.name}_periods')
        else:
            indexed = getattr(self._model, arg.name.split("[")[0]) # here, arg is single value so must find full IndexedVariable from whence it came
        
        node = args[0]
        iloc = self._model.periods.ord(self._index)
        if iloc == 1: # if the period is the first period, there is no prior period
            node = self.mean(indexed[self._model.periods.at(iloc)], 0)
        else:
            node = self.mean(indexed[self._model.periods.at(iloc)], indexed[self._model.periods.at(iloc - 1)])  # for every period other than the first period, get the prior period
                    
        return node

def total(expr, *args, **kwargs):
    model = model_finder(expr)
    return TotalExpression(expr, model, *args, **kwargs)

class TotalExpression(MethodExpression):
    PRECEDENCE = 20

    def create_node_with_local_data(self, args, classtype=None):
        """
        Returns cumulative summation of the indexed variable up to the period specified.
        """
        assert classtype is None

        node = args[0]
        indexed = getattr(self._model, node.name.split("[")[0]) 
        period = self._index[-1]
        nidxs = len(self._index) - 1
        sliced = (slice(None, None, None),)*nidxs

        return sum(v for v in indexed[sliced, period])

class PeriodizeVisitor(EXPR.ExpressionReplacementVisitor):
    """
    Transforms expressions with ScalarVar to expression using thei IndexedVar / Param derivatives
    """
    def __init__(self, model, index):
        super().__init__()
        self.model = model
        self.index = index

    def periodize_args(self, args):
        for arg in args:
            if type(arg) not in nonpyomo_leaf_types and arg.is_component_type() and not arg.is_indexed():
                if hasattr(self.model, f'{arg.name}_periods'):
                    indexed = getattr(self.model, f'{arg.name}_periods')
                    if indexed.dim() == 1: # if component only has 1 dimension, then that dimension should be the period
                        yield indexed[self.index[-1]]
                    else:
                        yield indexed[self.index]
                else:
                    scalar = getattr(self.model, f'{arg.name}') # have to swap out the AbstractModel Param for the Concrete Param
                    yield scalar
            else:
                yield arg
        
    def enterNode(self, node):
        args = list(self.periodize_args(node.args))
        return args, [False, args]

    def acceptChildResult(self, node, data, child_result, child_idx):
        if data[1][child_idx] is not child_result:
            data[1][child_idx] = child_result
            data[0] = True
        return data
    
    def exitNode(self, node, data):
        if hasattr(node, 'set_index'):
            node.set_index(self.index)
            node.set_model(self.model)  # backdoor to update expression during model construction ... without, certain parameters are not initialized by the time expressino is evauluated
        return node.create_node_with_local_data(tuple(data[1]))

class Where:
    """
    Handler for 
    """
    def __init__(self, mngr, tgt_name, inequal=None, big_M=1000):
        self.M = mngr
        self.short_name = tgt_name
        self.big_M = big_M

        if inequal is not None:
            self.parse_inequal(inequal)

    @property
    def target(self):
        return f'{self.short_name}_periods'

    @property
    def delta(self):
        return f'delta_{self.short_name}_periods'
    
    @property
    def comp(self):
        return f'{self._comp.name}_periods'

    @property
    def op(self):
        return self._op

    @property
    def c1(self):
        return f'{self.short_name}_constraints1'

    @property
    def c2(self):
        return f'{self.short_name}_constraints2'

    @property
    def c3(self):
        return f'{self.short_name}_constraints3'

    @property
    def constraints(self):
        return self.c1, self.c2, self.c3

    def parse_inequal(self, inequal):
        if not isinstance(inequal, InequalityExpression):
            raise ValueError('`where` can only accept `InequalityExpression` objects of form `x<0,x<=0,x>0,x>=0`')

        if inequal.strict:
            raise ValueError("currently `where` does not support 'strict' inequalities")

        _l, _r = inequal.args
        self._comp = _l if isinstance(_r, int) else _r
        self._op = 'le' if isinstance(_r, int)  else 'ge'
        self._inequal = inequal

    def ge_exprs(self, period):
        return {
            'c1': self.target[period] - self.big_M * self.delta[period] <= 0,
            'c2': self.comp[period] - self.target[period] <= 0,
            'c3': self.comp[period] - self.target[period] + self.big_M * self.delta[period] <= self.big_M
        }

    def ge_rules(self):
        return [ge_r1, ge_r2, ge_r3]

    def le_exprs(self, period):
        return {
            'c1': self.target[period] - self.big_M * self.delta[period] <= 0,
            'c2': -self.target[period] - self.comp[period] <= 0,
            'c3': self.comp[period] - self.target[period] + self.big_M * self.delta[period] <= self.big_M
        }

    def le_rules(self):
        return [le_r1, le_r2, le_r3]

    def exprs(self, period):
        if self.op == 'ge':
            return self.ge_exprs(period)
        elif self.op == 'le':
            return self.le_exprs(period)

    def rules(self):
        if self.op == 'ge':
            return self.ge_rules()
        elif self.op == 'le':
            return self.le_rules()

    def assign_vars(self):
        self.M.setvar(self.short_name) # root attribute i  s a shadow used to generate expressions; must be Var to allow for expressions
        self.M.setvar( # assign the IndexedParam; this contains actual data and is used in expression transformation
            f'{self.short_name}_periods',
            self.M.abstract.periods,
            domain=pyo.NonNegativeReals
        )
        self.M.setvar(f'delta_{self.short_name}') # root attribute i  s a shadow used to generate expressions; must be Var to allow for expressions
        self.M.setvar( # assign the IndexedParam; this contains actual data and is used in expression transformation
            f'delta_{self.short_name}_periods',
            self.M.abstract.periods,
            domain=pyo.Binary
        )

    def assign_exprs(self):
        for constr, rule in zip(self.constraints, self.rules()):
            rule = rule_factory(rule, self.target, self.comp, self.delta, self.big_M)
            setattr(self.M.abstract, constr, pyo.Constraint(self.M.abstract.periods, rule=rule))

def ge_r1(M, p, tgtstr, compstr, deltastr, big_M):
    target = getattr(M, tgtstr)
    delta = getattr(M, deltastr)
    return target[p] - big_M * delta[p] <= 0

def ge_r2(M, p, tgtstr, compstr, deltastr, big_M):
    target = getattr(M, tgtstr)
    comp = getattr(M, compstr)
    return comp[p] - target[p] <= 0

def ge_r3(M, p, tgtstr, compstr, deltastr, big_M):
    target = getattr(M, tgtstr)
    comp = getattr(M, compstr)
    delta = getattr(M, deltastr)
    return comp[p] - target[p] + big_M * delta[p] <= big_M

le_r1 = ge_r1

def le_r2(M, p, tgtstr, compstr, deltastr, big_M):
    target = getattr(M, tgtstr)
    comp = getattr(M, compstr)
    return -target[p] - comp[p] <= 0

le_r3 = ge_r3

def periodize(M, idx, tgtstr, expr):
    tgt = getattr(M, f'{tgtstr}_periods')
    walker = PeriodizeVisitor(M, idx)
    expr = walker.walk_expression(expr)
    return tgt[idx] == expr

def total(M, p, tgtstr):
    base = getattr(M, f'{tgtstr}_periods')
    tgt = getattr(M, f'tot_{tgtstr}_periods')
    
    sliced = (slice(None, None, None),)*(base.dim() - 1)
    return tgt[p] == sum(base[sliced, p])

def rule_factory(func, *args):
    @ft.wraps(func)
    def wrapper(M, *idx_args):
        return func(M, idx_args, *args)
    return wrapper

class obj_factory:
    def __init__(self, func, target=None):
        ft.update_wrapper(self, func)
        self.func = func
        self.target= target

    def __call__(self, *args):
        return self.func(*args, target=self.target)

class ModelManager:
    def __init__(self, stat, abstract=None):
        self.stat = stat
        if abstract is None:
            self._init_model()
        else:
            self.abstract = abstract
            
        self.solver = pyo.SolverFactory('glpk')

    def __getattribute__(self, name:str) -> typ.Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError as e:
            try:
                model = object.__getattribute__(self, '_model')
                return object.__getattribute__(model, name)
            except AttributeError:
                pass

            raise e

    @property
    def obj_assigned(self):
        return hasattr(self.abstract, 'obj')

    @property
    def has_instance(self):
        return hasattr(self, '_instance')

    @property
    def periods(self):
        return self.get('periods')

    def update_stat(self, stat):
        self.stat = stat

    def get(self, name):
        if self.has_instance:
            return getattr(self.instance, name)
        else:
            return getattr(self.abstract, name)

    def instance_data(self):
        root = {None: {}}
        data = root[None]
        data['periods'] = self.stat.periods.strftime(PERIODS_TO_STRF[self.stat.periods.freqstr]).tolist()
        
        if hasattr(self.abstract, 'multi'):
            aligned = self.stat.accounts.align(self.stat.by, with_item=True).strf_columns().drop_item_level()
            self._grouped_index = aligned[0].index
            self._has_multi = True
            data['multi'] = self._grouped_index.values.tolist()
        else:
            self._has_multi = False

        for var in self.abstract.component_objects(pyo.Param):
            name = var.name.split('_periods')[0]
            node = self.stat.G.nodes[name]
            if isinstance(var, IndexedParam):
                if node['is_multi']:
                    obj = aligned[name]
                    vardata = {}
                    for grouping, row in obj.iterrows():
                        grouping = (grouping, ) if isinstance(grouping, (str, int, float)) else grouping # handle single category groupbys
                        for period, val in row.iteritems():
                            vardata[(*grouping, period)] = val
                    data[var.name] = vardata
                else:
                    node = self.stat.G.nodes[name]
                    obj = node['obj']() if callable(node['obj']) else node['obj']
                    data[var.name] = {idx: v for idx, v in zip(data['periods'], obj)}
            else:
                if isinstance(var, ScalarParam) and node['nodetype'] == 'factor':
                    data[var.name] = {None: node['obj']}
                elif isinstance(var, ScalarParam) and node['nodetype'] == 'account':
                    data[var.name] = {None: 0}
                
        return root

    def _init_model(self):
        self.abstract = pyo.AbstractModel()
        self.abstract.periods = pyo.Set()

    def set_multi_index(self):
        if not self.has_multi_index():
            self.abstract.multi = pyo.Set()

    def has_multi_index(self):
        return hasattr(self.abstract, 'multi')

    def index_args(self, is_multi:bool):
        if is_multi:
            args = (self.abstract.multi, self.abstract.periods)
        else:
            args = (self.abstract.periods,)
        return args

    def setparam(self, name, *args, is_multi=False, mutable:bool=True, **kwargs):
        if '_periods' in name:
            args = (*self.index_args(is_multi), *args)
        setattr(self.abstract, name, pyo.Param(*args, **kwargs, mutable=mutable, domain=pyo.Reals))

    def setvar(self, name, *args, domain=pyo.Reals, **kwargs):
        setattr(self.abstract, name, pyo.Var(*args, **kwargs, domain=domain))

    def periodize(self, tgt, expr, is_multi:bool=False):
        rule = rule_factory(periodize, tgt, expr)
        setattr(self.abstract, f'{tgt}_constraints', pyo.Constraint(*self.index_args(is_multi), rule=rule))

    def total(self, comp):
        rule = rule_factory(total, comp.name)
        setattr(self.abstract, f'tot_{comp.name}_constraints', pyo.Constraint(self.abstract.periods, rule=rule))

    def where(self, *args, **kwargs):
        return Where(self, *args, **kwargs)

    def assign_vars(self, short_name, is_multi:bool=False):
        self.setvar(short_name) # root attribute is a shadow used to generate expressions; must be Var to allow for expressions
        self.setvar(f'{short_name}_periods', *self.index_args(is_multi))

    def assign_objective(self, target:str=None):
        if target is None:
            target = next(self._iter_indexed_vars())
        else:
            target = getattr(self.abstract, f'{target}_periods')

        if hasattr(self.abstract, 'obj'):
            self.abstract.del_component('obj')

        def rule(M, *args, target=None):
            if isinstance(target, str):
                target = getattr(M, target)
            return pyo.summation(target)

        rule = obj_factory(rule, target=target)
        self.abstract.obj = pyo.Objective(rule=rule, sense=pyo.maximize)

    def _update_param(self, name, key, value):
        param = getattr(self.instance, f'{name}_periods') ### Updates happen on instantiated models

        if isinstance(key, (str, int, float)):
            param[key] = value
        else:
            for k, v in zip(key, value):
                param[k] = v

    def _iter_indexed_vars(self, instance=False):
        model = self.instance if instance else self.abstract
        for var in model.component_objects(pyo.Var):
            if isinstance(var, IndexedVar):
                yield var

    def _iter_indexed_params(self):
        for var in self.instance.component_objects(pyo.Param):
            if isinstance(var, IndexedParam):
                yield var

    def create_instance(self):
        self.stat.close_model_context()
        data = self.instance_data()
        self.stat.open_model_context()
        self.instance = self.abstract.create_instance(data)

    def solve(self):
        from .elements.elements import index_values
        self.results = self.solver.solve(self.instance)

        msg = 'Your model does not have an optimal solution.'
        msg += ' See stat.M.results and stat.M.model.pprint to debug.'
        assert self.results.solver.termination_condition is TerminationCondition.optimal, msg

        for var in self._iter_indexed_vars(instance=True):
            node = self.stat.graph.nodes[var.name.split('_periods')[0]]
            values = np.array([pyo.value(var[idx]) for idx in var.keys()])
            if node['is_multi']:
                values = values.reshape(-1, self.stat.periods.size)
                node['obj'] = MultiLineItem(values, index=self._grouped_index, columns=self.stat.periods, short_name=node['short_name'], name=node['name'], graph=self.stat.graph)            
            elif node['is_total'] or self._has_multi:
                values = values.reshape(-1, self.stat.periods.size)
                index = index_values(self._grouped_index, placeholder='---')
                if isinstance(index, str):
                    index = (index, ) # if only one level, must make a tuple
                index = pd.MultiIndex.from_tuples((index,), names=self._grouped_index.names)
                node['obj'] = MultiLineItem(values, index=index, columns=self.stat.periods, short_name=node['short_name'], name=node['name'], graph=self.stat.graph)            
            else:
                node['obj'] = LineItem(values, index=self.stat.periods, short_name=node['short_name'], name=node['name'], graph=self.stat.graph)            
             
        return self.results



from .optimize import ModelManager

ScheduleFunction = None
########### Since financial statements are built from income statement to balance sheet, never nee dalst, only need `sum`


### WHEN RESAMPLE ###
# 1. Underlying DiGraph is copied to the new global statement constructor
# 2. This carries over all elements from the original statement
# 3. All elements containing a DiGraph MUST have their DiGraphs updated
# 4. The original instantiated ModelManager, however, is carried across in the copied DiGraph
# 5. After instantiation, the ModelManager must be overwritten, this time with the new grouped statement and a clone of the origina AbstractModel
# 5. So, ALL elements have a new DiGraph, which is the same among all the elements
# 6. And ALL elements have a new ModelManager and new underlying AbstractModel, which is a clone of the original
# 7. this should allow parameters to be updated via `instance_data` and new instances to be created without issue
class Grouper:
    def __init__(self, stat, *args, **kwargs):
        self.grouped = stat._constructor(name=stat.name, short_name=stat.short_name, graph=stat.G._graph.copy(), *args, **kwargs)
        for substat in stat.statements:
            self.grouped.add_statement(name=substat.name, short_name=substat.short_name) # have to create new substatement nodes
        self.grouped.G.graph['model_mngr'] = ModelManager(self.grouped, self.grouped.G.graph['model_mngr'].abstract.clone())
        self.grouped.elements.update_graphs(self.grouped.graph._graph)
        self.grouped.M.update_stat(self.grouped)
    
    def sum(self, *args, **kwargs):
        self.funcstr = 'sum'
        return self._call(*args, **kwargs)

    def last(self, *args, **kwargs):
        self.funcstr = 'sum'
        return self._call(*args, **kwargs)

    def mean(self, *args, **kwargs):
        self.funcstr = 'mean'
        return self._call(*args, **kwargs)

class FinStatResampler(Grouper):
    def __init__(self, stat, freq='A-DEC'):
        super().__init__(stat)
        self.freqstr = freq

        _new_periods = stat.periods.asfreq(self.freqstr).unique()
        self.grouped.update_periods(_new_periods)

    def resamp_accounts(self, *args, **kwargs):
        last = kwargs.pop('last', [])
        mean = kwargs.pop('mean', [])
        summ = kwargs.pop('summ', []) 
        
        for name, obj in self.grouped.G.filter_nodes_by_attribute('nodetype', 'account', data='obj', return_tuple=True):
            resampled = obj.resample(self.freqstr)
            if name in last:
                funcstr = 'last'
            elif name in mean:
                funcstr = 'mean'
            elif name in summ:
                funcstr = 'sum'
            else:
                funcstr = self.funcstr
            self.grouped.G.nodes[name]['obj'] = getattr(resampled, funcstr)(*args, **kwargs)

    def _call(self, *args, last=[], **kwargs):
        self.resamp_accounts(*args, last=last, **kwargs)
        
        if hasattr(self.grouped.M.abstract, 'obj'):
            self.grouped.M.create_instance()
            self.grouped.solve_out_of_context()
        
        return self.grouped

class FinStatGroupBy(Grouper):
    def __init__(self, stat, by):
        super().__init__(stat)
        self.grouped.G.graph['by'] = by

    def _call(self, *args, **kwargs):
        self.grouped.M.create_instance()
        self.grouped.solve_out_of_context()
        return self.grouped



import warnings
import typing as typ
import numpy as np
import pandas as pd

ElementsType = typ.TypeVar('ElementsType', bound='Elements')

def index_fillers(common_idx):
    fillers = []
    for i, col in common_idx.to_frame().iteritems():
        if pd.api.types.is_period_dtype(col):
            fillers.append(pd.Period('1900-1-1', freq='d'))
        elif pd.api.types.is_numeric_dtype(col):
            fillers.append(0)
        else:
            fillers.append('---')
    return fillers

def index_values(common_idx, placeholder='Other'):
    """
    Creates index values for 'Other' row created for MultiLineItem objects that are created via
    a MetricFunction or concating a set of LineItem objects.

    Index value is '---' or np.nan depending on dtype of the index value.

    Parameters
    -----------
    common_idx: pd.Index
        Index resulting from call to `common_level_values` method of Elements object
    """
    fillers = index_fillers(common_idx)
    row = tuple([placeholder] + fillers[1:]) if len(fillers) > 1 else placeholder

    return row

class Align:
    def __init__(self, items):
        self.items = items
        self.frames = self.items.idx_frames()
    
    def auto_groupby(self):
        shared = self.find_shared_levels()
        rank = self.level_rank(shared)
        match = self.match_rank(rank, self.items)
        return match[match > 0].index.tolist()
    
    def find_shared_levels(self):
        unq, counts = np.unique(np.concatenate(self.items._index_names), return_counts=True)
        return unq[counts == self.items.size]        
    
    def level_rank(self, shared):
        count = pd.Series([], dtype=float)
        for col in shared:
            unqlst = [idx[col].unique() for idx in self.frames]
            unq, counts = np.unique(np.concatenate(unqlst), return_counts=True)
            count.loc[col] = (counts >= self.items.size).sum() / unq.size

        return count.sort_values(ascending=False)    
    
    def match_rank(self, rank, items):
        frames = items.idx_frames()
        match_counts = pd.Series([], dtype='int')
        for i in range(1, rank.index.size):
            shared = rank.index[:-i]
            first, rest = frames[0], frames[1:]
            first = first[shared]

            with warnings.catch_warnings():
                warnings.simplefilter('ignore', FutureWarning) # catches warning about not inferring number types going forward

                if rest:
                    for nxt in rest:
                        nxt = nxt[shared]
                        first, match_count = self.merge_index_w_indicator(first, nxt)

                    match_counts.loc[shared[-1]] = match_count
    
        return match_counts.iloc[::-1]
    
    def merge_index_w_indicator(self, first, nxt):
        merged = pd.merge(first, nxt, how='outer', indicator=True)
        first = merged[merged._merge == 'both'].drop(columns='_merge')
        match_count = merged._merge.value_counts().both
        return first, match_count
    
    def common_index(self, frames):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', FutureWarning) # catches warning about not inferring number types going forward
        
            first, rest = frames[0], frames[1:]
        
            if rest:
                for nxt in rest:
                    if nxt.shape[0] > 1: # should exclude a total rows
                        first, _ = self.merge_index_w_indicator(first, nxt)

            common = first
            common = pd.MultiIndex.from_frame(common)

        if common.nlevels == 1:
            common = common.get_level_values(0)
            
        return common
    
    def append_uncommon(self, acct, common):
        uncommon = acct.loc[acct.index.difference(common)]
    
        if not uncommon.empty and acct.shape[0] > 1: # should exclude total rows
            other_row = uncommon.sum().rename(index_values(common))
            acct = acct.loc[common].sort_index()
            acct.loc[other_row.name, :] = other_row.values

        return acct
    
    def __call__(self, by=None, with_item=False):
        self.by = by if by else self.auto_groupby()
        
        grouped = self.items.groupby(self.by).drop_item_level()
        common = self.common_index(grouped.idx_frames()) 
        
        aligned = [] 
        for acct in grouped:
            acct = self.append_uncommon(acct, common)
            aligned.append(acct)
        
        items = Elements(*aligned)
        if with_item:
            return items.insert_item_level()
        else:
            return items

class concat:
    """
    Object-based concat method for Elements object. Handles different instances and
    combinations of LineItem objects and MultiLineItem objects.
    """
    def __init__(self, lineitems:ElementsType):
        self.lineitems = lineitems
    
    def _account_only_concat(self):
        concatted = pd.concat(self.lineitems, axis=1).T
        concatted.index.name = 'Item'
        concatted.index = pd.CategoricalIndex(
            concatted.index, 
            categories=self.lineitems.names, 
            ordered=True, 
            name='Item'
        )
        return concatted

    def _multi_level_concat(self, by:typ.Iterable=None, other=True):
        grouped = self.lineitems.align(by, with_item=True)
        concatted = pd.concat(grouped)
        concatted.categorize_level('Item', self.lineitems.names)
        return concatted
    
    def __call__(self, *args, **kwargs):
        if self.lineitems.is_multi.any():
            if self.lineitems.size > 1:
                concatted = self._multi_level_concat(*args, **kwargs) 
            else:
                concatted = self.lineitems[0] # if only 1 lineitem, just return that
        else:
            concatted = self._account_only_concat()

        return concatted

class Elements(np.ndarray):
    """
    Container for LineItem objects. Standard numpy methods and attributes are available.

    Includes several helper methods for manipulating sets of LineItem objects used in 
    MetricFunction and FinancialStatment.

    Issues arise from assigning LineItem objects directly to `obj`, so the str representation is instead.
    The underlying items manipulated are still the LineItem objects, given changes to __getitem__

    Parameters:
        elements: iterable of LineItem objects
    """

    def __new__(cls, *elements):
        obj = np.empty(len(elements), dtype='object').view(cls)
        obj[:] = elements

        for e in obj:
            setattr(obj, e.short_name, e)

        obj.concat = concat(obj)

        return obj
    
    def __array_finalize__(self, obj):
        if obj is None: return
        self.concat = getattr(obj, 'concat', None)

    def __repr__(self):
        joined = ', '.join(self.names)
        return f'Elements([{joined}])'

    def __add__(self, elements):
        return Elements(*[acct for acct in self] + [acct for acct in elements])

    @property
    def _index_names(self):
        return [np.array(acct.index.names) for acct in self]

    @property
    def names(self):
        return np.array([e.name for e in self])

    @property
    def short_names(self):
        return np.array([e.short_name for e in self])

    @property
    def is_multi(self):
        return np.array([e.is_multi for e in self])

    @property
    def any_multi(self):
        return np.any(self.is_multi)

    @property
    def are_aligned(self):
        return all([self[0].index.equals(e.index) for e in self[1:]])

    @property
    def have_single_row(self):
        return np.array([e.has_single_row for e in self])

    def update_graphs(self, graph):
        [e.update_graph(graph) for e in self]

    def idx_frames(self):
        return [e.index.to_frame().reset_index(drop=True).apply(lambda col: col.astype(object)) for e in self]

    def groupby(self, by):
        return Elements(*[e.groupby(by).sum() for e in self])

    def strf_columns(self):
        from ..base import PERIODS_TO_STRF
        items = []
        for item in self:
            item.columns = item.columns.strftime(PERIODS_TO_STRF[item.columns.freqstr]).tolist()
            items.append(item)
        return Elements(*items)

    def to_multi(self):
        return Elements(*[e.to_multi() if not is_multi else e for is_multi, e in zip(self.is_multi, self)])

    def to_elements(self):
        return Elements(*[e.to_account() if single else e for single, e in zip(self.have_single_row, self)])

    def drop_item_level(self):
        return Elements(*[e.droplevel('Item') if 'Item' in e.index.names else e for e in self])

    def insert_item_level(self):    
        return Elements(*[e.insert_item_level() for e in self])
    
    def by_position(self):
       return Elements(*sorted(self, key=lambda n: n.graph.nodes[n.short_name]['position']))

    @property
    def align(self):
        return Align(self)
        